/*
 * Copyright 1999-2009 University of Chicago
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef GLOBUS_DONT_DOCUMENT_INTERNAL
/**
 * @file module.c
 * GSSAPI module activation code
 *
 * $RCSfile: esg_gridmap_assist.c,v $
 * $Revision: 1.25 $
 * $Date $
 */
#endif

#include "globus_common.h"
#include "globus_gsi_authz.h"
#include "globus_gsi_authz_callout_error.h"
#include "globus_gsi_system_config.h"
#include "gaa.h"
#include "gaa_plugin.h"
#include "gaa_gss_generic.h"
#include "gssapi.h"
#include "globus_gss_assist.h"
#include "globus_gridmap_callout_error.h"
#include "esg_gridmap_assist.h"

#include "openssl/ssl.h"
#include "openssl/err.h"
#include "openssl/x509.h"

/* static const gss_OID_desc esg_role_extension_oid = */
/*     {10, "\x01\x02\x03\x04\x04\x03\x02\x01\x07\x08"}; */

static const gss_OID_desc esg_role_extension_oid =
    {9, "\x2a\x03\x04\x04\x03\x02\x01\x07\x08"};
const gss_OID_desc *const esg_role_extension = 
    &esg_role_extension_oid;

static gss_ctx_id_t *s_context = NULL;


void set_security_context(gss_ctx_id_t *context)
{
    s_context = context;
}

/* returns 0 on success; -1 on failure */
int parse_config_file(char *config_file, config_t *config)
{
    int ret = -1;
    if (config_file && config)
    {
        struct stat buf;
        FILE *fconfig = NULL; 
        char *ptr = NULL, *dest = NULL;
        char line[512] = {0};
        char key[MAX_CONFIG_KV_LEN] = {0}, value[MAX_CONFIG_KV_LEN] = {0};

        memset(&buf, 0, sizeof(struct stat));
        if (stat(config_file, &buf) != 0)
        {
            config_file = getenv("ESG_GRIDMAP_ASSIST_CONFIG_FILE");
            if (stat(config_file, &buf) != 0)
            {
                fprintf(stderr, "ERROR: No configuration file found!\n");
                goto end;
            }
        }

/*         fprintf(stderr, "*** USING CONFIG FILE %s\n", config_file); */
        fconfig = fopen(config_file, "r");
        if (fconfig)
        {
            while (fgets(line, 512, fconfig) != NULL)
            {
                if ((line[0] == '#') || (strlen(line) < 2))
                {
                    continue;
                }
                else
                {
                    memset(key, 0, MAX_CONFIG_KV_LEN);
                    memset(value, 0, MAX_CONFIG_KV_LEN);

                    for(ptr = line, dest = key; *ptr != '\0'; ptr++)
                    {
                        if (*ptr == '=')
                        {
                            dest = value;
                        }
                        else if ((*ptr == ' ') || (*ptr == '\t') || (*ptr == '\r') || (*ptr == '\n'))
                        {
                            continue;
                        }
                        else
                        {
                            *dest++ = *ptr;
                        }
                    }
                    if (strcasecmp(key, "attribute") == 0)
                    {
                        strncpy(config->admin_attr,
                                value, MIN(strlen(value), MAX_CONFIG_KV_LEN));
                    }
                    else if (strcasecmp(key, "localaccount") == 0)
                    {
                        strncpy(config->admin_local_account,
                                value, MIN(strlen(value), MAX_CONFIG_KV_LEN));
                    }
                    else
                    {
                        fprintf(stderr, "Unrecognized key found: %s\n", key);
                    }
                }
            }
            fclose(fconfig);
        }
    }

    ret = ((strlen(config->admin_attr) && strlen(config->admin_local_account)) ? 0 : 1);

  end:
    return ret;
}

static char *get_error_msg(int ret)
{
    return globus_object_printable_to_string(
        globus_error_get((globus_result_t)ret));
}

/*
  returns 0 if the extension assertion exists and the config object is properly populated
  returns 1 if there is no extension or assertion present
  returns -1 on error
*/
int parse_credential_extension(config_t *config)
{
    int ret = -1, i = 0;
    OM_uint32 minor_status, r;
    char *assertion = NULL;
    gss_buffer_set_t data_set = 0;
    globus_size_t assertion_len;
    globus_bool_t der_encoded_assertion = FALSE;

    if (config)
    {
        r = gss_create_empty_buffer_set(&minor_status, &data_set);
        if (r)
        {
            fprintf(stderr, "gss_create_empty_buffer_set failed: (%s)\n",
                    get_error_msg(minor_status));
            goto end;
        }

        r = gss_inquire_sec_context_by_oid(
            &minor_status, *s_context, (gss_OID)esg_role_extension, &data_set);
        if (r == GSS_S_COMPLETE)
        {
            der_encoded_assertion = GLOBUS_TRUE;
        }

        if (data_set == NULL)
        {
            fprintf(stderr, "gss_inquire_sec_context_by_oid returned NULL: (%s)\n",
                    get_error_msg(minor_status));
            goto end;
        }

        if (data_set->count == 0)
        {
            fprintf(stderr, "No extension found: (%s)\n",
                    get_error_msg(minor_status));
            goto end;
        }

        for (i = 0; i < data_set->count; i++)
        {
            if (data_set->elements[i].length && data_set->elements[i].value)
            {
                assertion = malloc(data_set->elements[i].length+1);
                memcpy(assertion,
                       data_set->elements[i].value,
                       data_set->elements[i].length);
                assertion_len = data_set->elements[i].length;
                assertion[assertion_len] = '\0';
                break;
            }
        }

        if (!assertion)
        {
            fprintf(stderr, "No assertion found: (%s)\n",
                    get_error_msg(minor_status));
            goto end;
        }

        if (der_encoded_assertion)
        {
            char *encoded_assertion = assertion;
            char *decoded_assertion = NULL;
            ASN1_UTF8STRING *asn1_str;

            asn1_str = d2i_ASN1_UTF8STRING(
                NULL, &encoded_assertion, assertion_len);

            if (asn1_str)
            {
                decoded_assertion = malloc(asn1_str->length + 1);
                memcpy(decoded_assertion, asn1_str->data, asn1_str->length);
                decoded_assertion[asn1_str->length] = 0;

                ASN1_UTF8STRING_free(asn1_str);
                globus_free(assertion);
                assertion = decoded_assertion;
            }
            else
            {
                fprintf(stderr, "Improperly encoded assertion found: (%s)\n",
                        get_error_msg(minor_status));
                goto end;        
            }
        }

/*         fprintf(stderr, "*** GOT ASSERTION: %s\n", assertion); */

        /* parse out key attribute stored in config from assertion now */
        {
            int admin_attr_len = strlen(config->admin_attr);
            char *ptr = NULL;
            char *end = (char *)(assertion + (strlen(assertion) * sizeof(char)));
            for(ptr = assertion; (ptr && (ptr != end) && (*ptr != '\0')); ptr++)
            {
                if ((*ptr == config->admin_attr[0]) &&
                    ((ptr - end) >= strlen(ptr)))
                {
                    if (strncasecmp(ptr, config->admin_attr, admin_attr_len) == 0)
                    {
/*                         fprintf(stderr, "FOUND BDM ADMIN ATTRIBUTE!!! (at %s)\n", ptr); */
                        ret = 0;
                        break;
                    }
                }
            }
        }
    }

  end:

    if (data_set)
    {
        gss_release_buffer_set(&minor_status, &data_set);
    }
    if (assertion)
    {
        free(assertion);
    }
    return ret;
}
